#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#include <linux/dvb/dmx.h>
#include <linux/dvb/frontend.h>

#include "nitscan.h"

static int cont = 1;
static int channel = 13;

static int skip_1seg = 0;
static int exact_pid = 0;

static void stop_loop(int sig)
{
	cont = 0;
	return;
}

static int dump_si(struct isdbt_si *si)
{
	int i, j;
	int rc = 0;

#define MAX_PID 6
#define MAX_PIDLEN ((4+1)*MAX_PID)
	char pname[8], vpid[MAX_PIDLEN], apid[MAX_PIDLEN];

	struct pmt *pmt;
	char *vp, *ap;

	int prognum = 0;

	if (si->pat.ver == -1) {
		fprintf(stderr, " PAT not recived.\n");
		return -1;
	}

	if (si->nit.ver != -1)
		rc = si->nit.ts_info[0].rc_key_id;

	for (i=0; i<si->pat.num_prog; i++) {
		pmt = si->pat.prog[i].pmt;
		if (!pmt || pmt->ver == -1) {
			fprintf(stderr, " PMT for [%04hx] not recieved.\n",
				si->pat.prog[i].prog_id);
			continue;
		}

		if (!pmt->num_es) {
			fprintf(stderr, " prog[%04hx] has no Video/Audio ES.\n",
				pmt->prog_id);
			continue;
		}

/*
		if (si->nit.ver != -1) {
			for (j=0; j<si->nit.ts_info[0].num_service; j++) {
				if (pmt->prog_id == si->nit.ts_info[0].services[j].service_id)
					break;
			}
			if (j < si->nit.ts_info[0].num_service && 
				si->nit.ts_info[0].services[j].partial) {
				fprintf(stderr, " prog[%04hx] is for partial receving.\n", pmt->prog_id);
				continue;
			}
		}
 */

		if (rc)
			snprintf(pname, sizeof(pname), "%02d%1d", rc, prognum++);
		else
			snprintf(pname, sizeof(pname), "[%04hx]", pmt->prog_id);

		vp = vpid;
		ap = apid;
		for (j=0; j<pmt->num_es; j++) {
			if (pmt->es[j].es_type == ES_TYPE_MPEG2 ||
			    (!skip_1seg && pmt->es[j].es_type == ES_TYPE_H264)) {
				if (vp + 5 >= vpid + sizeof(vpid)) {
					fprintf(stderr, " vpid overflowed.\n");
					break;
				}
				if (vp != vpid)
					*vp++ = '+';
				vp += snprintf(vp, 5, "%hu", pmt->es[j].es_pid);
			} else if (pmt->es[j].es_type == ES_TYPE_AAC) {
				if (ap + 5 >= apid + sizeof(apid)) {
					fprintf(stderr, " apid overflowed.\n");
					break;
				}
				if (ap != apid)
					*ap++ = '+';
				ap += snprintf(ap, 5, "%hu", pmt->es[j].es_pid);
			}
		}

		if (vp == vpid) continue;

		if (vp + 7 < vpid + sizeof(vpid))
			vp += snprintf(vp, 8, "+0+%hu", si->pat.prog[i].pmt_pid);
		if (vp + 5 < vpid + sizeof(vpid) && pmt->pcr_pid != 0x1fff)
			vp += snprintf(vp, 6, "+%hu", pmt->pcr_pid);

		if (vp + 5 < vpid + sizeof(vpid) && pmt->prog_ecm.ca_pid != 0x1fff)
			vp += snprintf(vp, 6, "+%hu", pmt->prog_ecm.ca_pid);

		*vp = '\0';
		*ap = '\0';

		printf("%s:%d:INVERSION_AUTO:BANDWIDTH_6_MHZ:FEC_7_8:FEC_1_2:"
		       "QAM_64:TRANSMISSION_MODE_2K:GUARD_INTERVAL_1_16:"
		       "HIERARCHY_NONE:%s:%s:%d\n",
		       pname,
		       (473 + 6 * (channel - 13))*1000000,
		       exact_pid ? vpid : "8192",
		       exact_pid ? apid : "",
		       pmt->prog_id);
	}

	return 0;
}


static const char *usage = "\n"
	"usage %s [options...] {-c | <channel-list-file>}\n"
	"	-c N	scan only on physical channel N (13..62)\n"
	"	-a N	use /dev/dvb/adapterN\n"
	"	-f N	use /dev/dvb/adapter?/frontendN\n"
	"	-d N	use /dev/dvb/adapter?/demuxN\n"
	"	-p	output the PIDs of each program, instead of wildcard\n"
	"	-s	skip H264(\"One Seg.\") programs\n";


int main(int argc, char **argv)
{

	int adapter = 0, frontend = 0, demux = 0;
	char fname[32], dname[32];
	int fe, fd, cfd;
	struct dvb_frontend_parameters param;
	struct dmx_pes_filter_params flt;
	fe_status_t stat;
	struct sigaction act;
	struct itimerval it, itz;
	struct isdbt_si si;
	int opt, i, c, ret;
	size_t len;
	uint8_t buf[188];
	uint8_t *p;
	uint16_t pid;
	int ch_list[50];
	int ch_len = 0;
	char linebuf[80];
	FILE *fc;

	if (argc < 2) {
		fprintf(stderr, usage, argv[0]);
		return -1;
	}

	while ((opt = getopt(argc, argv, "c:a:f:d:sp")) != -1) {
		switch (opt) {
		case 'a':
			adapter = strtoul(optarg, NULL, 0);
			break;
		case 'c':
			ch_list[0] = strtoul(optarg, NULL, 0);
			ch_len = 1;
			break;
		case 'd':
			demux = strtoul(optarg, NULL, 0);
			break;
		case 'f':
			frontend = strtoul(optarg, NULL, 0);
			break;
		case 'p':
			exact_pid = 1;
			break;
		case 's':
			skip_1seg = 1;
			break;
		default:
			fprintf(stderr, usage, argv[0]);
			return -1;
		};
	}

	if (optind == argc && !ch_len) {
		fprintf(stderr, usage, argv[0]);
		return -1;
	}

	if (optind < argc) {
		cfd = open(argv[optind], O_RDONLY);
		if (cfd < 0 || !(fc = fdopen(cfd, "r")) ) {
			perror(" failed to open the channel list file.");
			return -1;
		}
		ch_len = 0;
		while (ch_len < 50 && fgets(linebuf, sizeof(linebuf), fc)) {
			ch_list[ch_len] = strtoul(linebuf, NULL, 0);
			if (ch_list[ch_len] < 13 || ch_list[ch_len] > 62)
				continue;
			ch_len++;
		}
		fclose(fc);
	}
	fprintf(stderr, " scanning %d channel(s).\n", ch_len);


	snprintf(fname, sizeof(fname),
		"/dev/dvb/adapter%i/frontend%i", adapter, frontend);

	snprintf(dname, sizeof(dname),
		"/dev/dvb/adapter%i/demux%i", adapter, demux);


	fe = open(fname, O_RDWR);
	if (fe < 0) {
		perror(" FE open failed.");
		return -fe;
	}

	fd = open(dname, O_RDONLY);
	if (fd < 0) {
		perror(" demuxer open failed.");
		close(fe);
		return -fd;
	}
	ioctl(fd, DMX_SET_BUFFER_SIZE, 4 * 1024 * 1024);

	flt.pid = 0x2000;
	flt.input = DMX_IN_FRONTEND;
	flt.output = DMX_OUT_TAP;
	flt.pes_type = DMX_PES_OTHER;
	flt.flags = 0;

	if (ioctl(fd, DMX_SET_PES_FILTER, &flt) < 0) {
		perror("DMX_SET_PES_FILTER");
		return -fd;
	}

	sigemptyset(&act.sa_mask);
	act.sa_handler = stop_loop;
	act.sa_flags = 0;
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGALRM, &act, NULL);

	itz.it_interval.tv_sec = 0;
	itz.it_interval.tv_usec = 0;
	itz.it_value.tv_sec = 0;
	itz.it_value.tv_usec = 0;

	c = 0;

ch_loop:
	cont = 1;
	channel = ch_list[c++];

	ioctl(fd, DMX_STOP, 1);
	clean_si(&si);
	init_si(&si);

	usleep(500*1000);
	param.frequency = (473 + 6 * (channel - 13)) * 1000000;
	if (ioctl(fe, FE_SET_FRONTEND_TUNE_MODE, FE_TUNE_MODE_ONESHOT) ||
	    ioctl(fe, FE_SET_FRONTEND, &param) < 0) {
		fprintf(stderr, " FE set channel failed.\n");
		goto bailout;
	}
	fprintf(stderr, " scanning physical ch.%02d...\n", channel);
	usleep(500*1000);
	if (ioctl(fe, FE_READ_STATUS, &stat) < 0) {
		fprintf(stderr, " FE read status failed.\n");
		goto bailout;
	}
	if (!(stat & FE_HAS_LOCK)) {
		fprintf(stderr, " FE PLL not locked for ch:%d.\n", channel);
		goto bailout;
	}

	if (ioctl(fd, DMX_START, 1) < 0) {
		perror("DMX_START");
		goto bailout;
	}

	it.it_interval.tv_sec = 0; /* one shot */
	it.it_interval.tv_usec = 0;
	it.it_value.tv_sec = 5;
	it.it_value.tv_usec = 0;
	if (setitimer(ITIMER_REAL, &it, NULL) < 0) {
		perror("setitimer");
		goto bailout;
	}

	while (cont) {
		buf[0] = 0;
		if (read(fd, buf, 1) < 1) {
			fprintf(stderr, "Error: cannot read data.\n");
			break;
		}

		for (i=0; i< 188; i++) {
			if (buf[0] == 0x47)
				break;
			if (!(i % 10))
				fprintf(stderr, ".");
			usleep(100*1000);
		}
		if (i)
			fprintf(stderr, "\n");
		if (i == 188) {
			fprintf(stderr, "Error: cannot sync.\n");
			break;
		}

		p = &buf[1];
		do {
			len = read(fd, p, buf + 188 - p);
			if (len < 0) {
				perror(" read failed.");
				goto bailout;
			}
			p += len;
		} while (p < buf + 188);

		if (!(buf[3] & TS_F_PAYLOAD)) continue;

		pid = ((buf[1] & 0x1f) << 8) + buf[2];
		if (pid == PID_PAT) {
			doSection(buf, &si.pat.sec, doPAT, &si.pat);
			continue;
		} else if (pid == PID_NIT) {
			doSection(buf, &si.nit.sec, doNIT, &si.nit);
			continue;
		}

		for (i=0; si.pat.ver != -1 && i<si.pat.num_prog; i++) {
			if (!si.pat.prog[i].pmt)
				continue;

			if (pid == si.pat.prog[i].pmt_pid) {
				doSection(buf, &si.pat.prog[i].pmt->sec,
					  doPMT, si.pat.prog[i].pmt);
				break;
			}
/*
			if (pid == si.pat.prog[i].pmt->prog_ecm.cas_pid) {
				doSection(buf, &si.pat.prog[i].xxx, 
					  doECM, &si.pat.prog[i]);
				break;
			}
*/
		}

/*
		if (si.pat.ver != -1 && i < si.pat.num_prog) continue;
		for (i=0; si.cat.ver != -1 && i<si.cat.num_cas; i++) {
			if (pid == si.cat.emm[i].cas_pid) {
				doSection(buf, doEMM);
				break;
			}
		}
*/

		if (si.pat.ver != -1 && si.nit.ver != -1) {
			for (i=0; i<si.pat.num_prog; i++)
				if (si.pat.prog[i].pmt->ver == -1)
					break;
			if (i == si.pat.num_prog)
				break;
		}
	}

	ret = dump_si(&si);
	
bailout:
	setitimer(ITIMER_REAL, &itz, NULL);
	if (c < ch_len) goto ch_loop;

	ioctl(fd, DMX_STOP, 1);
	close(fd);
	close(fe);
	return ret;
}
