#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>

#include <linux/dvb/frontend.h>

static char FRONTEND_DEV[64];

static const char *CHANNEL_FILE = "/.jzap/channels.conf";
static const char *usage = "\nusage: %s [-a adapter_num] [-f frontend_id] [-c conf_file] <channel_name>\n\n";

static int
parse(const char *fname, const char *channel,
      struct dvb_frontend_parameters *frontend)
{
	FILE *f;
	char buf[256];
	char *p;
	int cmplen;
	unsigned long freq;

	if (!(f = fopen(fname, "r"))) {
		perror("could not open the conf-file.");
		return -1;
	}

	cmplen = strlen(channel);
	if (cmplen > sizeof(buf) - 1)
		cmplen = sizeof(buf) -1;

	while (fgets(buf, sizeof(buf), f)) {
		p = buf;
		if (strncmp(channel, strsep(&p, ":"), strlen(channel)))
			continue;

		freq = strtoul(p, NULL, 10);
		if (!freq) {
			fprintf(stderr, "Error: reading the channel frequency.\n");
			fclose(f);
			return -1;
		}
		frontend->frequency = freq;
		/* set non-used parameters. */
		/*  some may have wrong values, but don't care. */
		frontend->inversion = INVERSION_AUTO;
		frontend->u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
		frontend->u.ofdm.code_rate_HP = FEC_7_8;
		frontend->u.ofdm.code_rate_LP = FEC_1_2;
		frontend->u.ofdm.constellation = QAM_64;
		frontend->u.ofdm.transmission_mode = TRANSMISSION_MODE_2K;
		frontend->u.ofdm.guard_interval = GUARD_INTERVAL_1_16;
		frontend->u.ofdm.hierarchy_information = HIERARCHY_AUTO;

		fclose(f);
		return 0;
	}

	fprintf(stderr, "Error: could not find the channel name:'%s'.\n", channel);
	fclose(f);
	return -1;
}


int
main(int argc, char **argv)
{
	struct dvb_frontend_parameters frontend_param;
	fe_status_t status;

	char *homedir = getenv("HOME");
	char *confname = NULL;
	char *channel = NULL;
	int adapter = 0, frontend = 0;
	int frontend_fd;
	int opt;

	while ((opt=getopt(argc, argv, "hrn:a:f:d:c:")) != -1) {
		switch (opt) {
		case 'a':
			adapter = strtoul(optarg, NULL, 0);
			break;
		case 'f':
			frontend = strtoul(optarg, NULL, 0);
			break;
		case 'd':
		case 'r':
			break; // for the compatibility with {t,s,a}zap
		case 'c':
			confname = optarg;
			break;
		case '?':
		case 'h':
		default:
			fprintf(stderr, usage, argv[0]);
			return -1;
		};
	}

	if (optind < argc)
		channel = argv[optind];

	if (!channel) {
		fprintf(stderr, usage, argv[0]);
		return -1;
	}

	snprintf(FRONTEND_DEV, sizeof(FRONTEND_DEV),
		 "/dev/dvb/adapter%i/frontend%i", adapter, frontend);

	printf("using '%s'\n", FRONTEND_DEV);

	memset(&frontend_param, 0, sizeof(struct dvb_frontend_parameters));

	if (!confname) {
		if (!homedir)
			fprintf(stderr, "Error: $HOME not set.\n");
		confname = malloc(strlen(homedir) + strlen(CHANNEL_FILE) + 1);
		if (!confname) {
			fprintf(stderr, "Error: not enough memory.\n");
			return -1;
		}
		memcpy(confname, homedir, strlen(homedir));
		memcpy(confname + strlen(homedir), CHANNEL_FILE,
			strlen(CHANNEL_FILE) + 1);
	}

	if (parse(confname, channel, &frontend_param))
		return -1;

	if ((frontend_fd = open(FRONTEND_DEV, O_RDWR)) < 0) {
		perror("failed to opening the frontend device.");
		return -1;
	}

	printf("tuning to %i Hz...", frontend_param.frequency);
	if (ioctl(frontend_fd, FE_SET_FRONTEND_TUNE_MODE, FE_TUNE_MODE_ONESHOT) < 0) {
		perror("ioctl FE_SET_FRONTEND_TUNE_MODE failed.");
		close(frontend_fd);
		return -1;
	}
	if (ioctl(frontend_fd, FE_SET_FRONTEND, &frontend_param) < 0) {
		perror("ioctl FE_SET_FRONTEND failed.");
		close(frontend_fd);
		return -1;
	}
	usleep(500*1000);

	if (ioctl(frontend_fd, FE_READ_STATUS, &status) < 0) {
		perror("ioctl FE_READ_STATUS failed.");
		close(frontend_fd);
		return -1;
	}
	if (status & FE_HAS_LOCK) {
		printf("OK\n");
	} else {
		printf("NG\n");
		close(frontend_fd);
		return -2;
	}

	close(frontend_fd);
	return 0;
}
