#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>
#include <PCSC/pcsclite.h>

#define dprintf(args...) do { \
	if (!background) \
		fprintf(stderr, args); \
	} while (0)

#define xprintf(args...) do {\
	if (background) \
		syslog(LOG_WARNING, args); \
	else \
		fprintf(stderr, args); \
	} while (0)


/* example ATR of BCAS  note: byte[2],[4],[8],[9],[11],[12] may vary. */
static const BYTE bcas_atr[] = {
	0x3B, 0xF0, 0x12, 0x00, 0xFF, 0x91, 0x81,
	0xB1, 0x7C, 0x45, 0x1F, 0x03, 0x99
};
static const DWORD BCAS_ATR_LEN = sizeof(bcas_atr);

static int background = 0;

static int get_card(SCARDCONTEXT cxt, SCARDHANDLE *h, char *iccname)
{
	LONG ret;
	DWORD len, protocol;
	char *p, *readers;
	DWORD nlen, alen;
	DWORD state;
	BYTE atr[MAX_ATR_SIZE];

	ret = SCardListReaders(cxt, NULL, NULL, &len);
	if (ret != SCARD_S_SUCCESS) {
		xprintf("Cannot get length of the IC card list: %s\n",
			pcsc_stringify_error(ret));
		return 1;
	}

	readers = malloc(len);
	if (!readers) {
		xprintf("Cannot alloc mem for the IC card list. %m\n");
		return -1;
	}

	ret = SCardListReaders(cxt, NULL, readers, &len);
	if (ret != SCARD_S_SUCCESS) {
		xprintf("Cannot get card list. %s\n",
			pcsc_stringify_error(ret));
		free(readers);
		return 1;
	}

	p = readers;
	while (*p) {
		if (iccname && strlen(p) == strlen(iccname) && strncmp(p, iccname, strlen(p)))
				goto next_dev;

		ret = SCardConnect(cxt, p, SCARD_SHARE_SHARED,
				   SCARD_PROTOCOL_T1, h, &protocol);
		if (ret != SCARD_S_SUCCESS) {
			dprintf(" cannot connect to [%s]/%s.\n",
				p, pcsc_stringify_error(ret));
			goto next_dev;
		}
/*
		SCardReconnect(h, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1,
				SCARD_RESET_CARD, &protocol); 
		usleep(10*1000);
*/
		alen = sizeof(atr);
		ret = SCardStatus(*h, NULL, &nlen, &state, &protocol,
				  atr, &alen);
		dprintf(" state:0x%02hhx atr-len:%d\n", state, alen);

/*
		{ int i;
		for(i=0; i<alen; i++) dprintf(" %02hhx", atr[i]);
		dprintf("\n");
		}
 */

		if (ret == SCARD_S_SUCCESS && state & SCARD_PRESENT &&
		    alen == BCAS_ATR_LEN &&
		    atr[0] == bcas_atr[0] && atr[1] == bcas_atr[1]) {
			dprintf("found a (likely) BCAS card[%s].\n", p);
			free(readers);
			return 0;
		}
		SCardDisconnect(*h, SCARD_RESET_CARD);
	next_dev:
		p += strlen(p) + 1;
	}
	free(readers);

	xprintf("cannot find an available BCAS card.\n");
	*h = 0;
	return 2;
}

static int exit_loop = 0;

static void stop_loop(int sig)
{
	exit_loop = 1;
	return;
}

static void set_restart(int sig)
{
	exit_loop = 2;
	return;
}

static const char usage[] = "\nUsage:%s [-a <adapter#>] [-i <ICCR_name>] [-r <num>] [-b]\n"
"        -a <adapter#>  DVB adapter number to be used\n"
"        -i <ICC name>  IC card reader's name (as in pcsc_scan)\n"
"        -r <num>       retry until <num> consecutive erros\n"
"        -b             wait and retry on errors, don't exit.(daemon mode)\n";

int main(int argc, char **argv)
{
	int adapter = 0;
	char *iccname = NULL;
	int retry_count = 3, retry = 0;

	int opt, r, i, card_ok;
	char dname[32];
	struct sigaction act;

	LONG ret;
	SCARDCONTEXT cxt;
	SCARDHANDLE  hcard;
	DWORD protocol;
	DWORD rlen;
	SCARD_IO_REQUEST rpci;

	int ca_fd;
	unsigned char tbuf[256];
	unsigned char rbuf[256];
	unsigned char *p;
	int len;

	while ((opt = getopt(argc, argv, "a:i:r:b")) != -1) {
		switch (opt) {
		case 'a':
			adapter = strtoul(optarg, NULL, 0);
			break;
		case 'i':
			iccname = optarg;
			break;
		case 'r':
			retry_count = strtoul(optarg, NULL, 0);
			break;
		case 'b':
			background = 1;
			openlog(argv[0], LOG_PID | LOG_CONS, LOG_USER);
			break;
		default:
			if (background)
				syslog(LOG_ERR, " invalid options.\n");
			else
				fprintf(stderr, usage, argv[0]);
			return -1;
		}
	}

	ret = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &cxt);
	if (ret != SCARD_S_SUCCESS) {
		xprintf("Can't connect to pcscd. %s\n",
			pcsc_stringify_error(ret));
		return -1;
	}

	snprintf(dname, sizeof(dname), "/dev/dvb/adapter%i/ca0", adapter);
//	snprintf(dname, sizeof(dname), "testdat");


	sigemptyset(&act.sa_mask);
	act.sa_handler = stop_loop;
	act.sa_flags = 0;
	sigaction(SIGINT, &act, NULL);

repeat:
	ca_fd = open(dname, O_RDWR);
	if (ca_fd < 0) {
		xprintf("swbcas device open failed. %m\n");
		r = ca_fd;
		goto bailout;
	}

	card_ok = 0;
	while (!exit_loop) {
		if (!card_ok) {
			r = get_card(cxt, &hcard, iccname);
			if (r < 0)
				goto bailout;
			if (r > 0) {
				retry++;
				if (!background && retry > retry_count)
					goto bailout;
				sleep(5);
				continue;
			}
			card_ok = 1;
		}
				
		/* read-(IC Card)-write loop */
		/* read {len, seq, APDU[len]} */
		r = read(ca_fd, tbuf, sizeof(tbuf));
		if (r < 0 && errno == EINTR) {
			usleep(200 * 1000);
			continue;
		}
		if (r <= 0) {
			xprintf("dev read header failed. %m\n");
			break;
		}

		len = tbuf[0] + 2 - r;
		p = tbuf + r;
		while (len > 0) {
			r = read(ca_fd, p, sizeof(tbuf) - (p - tbuf));
			dprintf("additional read ret:%d\n", r);
			if (r <= 0) {
				xprintf("dev read data failed. %m\n");
				goto bailout;
			}
			p += r;
			len -= r;
		}
		dprintf("cmd[%d] read. -> ", p - tbuf);

		/* send to the card */
		memcpy(&rpci, SCARD_PCI_T1, sizeof(rpci));
		rlen = sizeof(rbuf) - 2;
/*
		{ int i;
		for(i=0; i < tbuf[0] + 2; i++) dprintf(" %02hhx", tbuf[i]);
		dprintf("\n");
		}
 */

		ret = SCardTransmit(hcard, SCARD_PCI_T1, tbuf + 2, tbuf[0],
				    &rpci, rbuf + 2, &rlen);
		if (ret == SCARD_W_RESET_CARD) {
			ret = SCardReconnect(hcard, SCARD_SHARE_SHARED,
					     SCARD_PROTOCOL_T1,
					     SCARD_LEAVE_CARD, &protocol);
		}

		/* write back to the dev */
		rbuf[1] = tbuf[1];
		if (ret == SCARD_S_SUCCESS) {
			rbuf[0] = rlen & 0xff;
		} else {
			xprintf("bcas transmit failed.%s\n",
				pcsc_stringify_error(ret));
			rbuf[0] = 0; /* for reporting error, send empty data */
			SCardDisconnect(hcard, SCARD_RESET_CARD);
			card_ok = 0;
			retry++;
			if (!background && retry > retry_count)
				exit_loop = 1;
		}

		dprintf("resp(%d) recved. -> ", rlen);

		r = write(ca_fd, rbuf, rbuf[0] + 2);
		if (r < 0 && errno == EINTR) {
			usleep(200*1000);
			continue;
		}
		if (r < rbuf[0] + 2) {
			xprintf("dev write error. %m\n");
			break;
		}
		dprintf("wrote back to the dev.\n");
/*
		{ int i;
		for(i=0; i < rbuf[0] + 2; i++) dprintf(" %02hhx", rbuf[i]);
		dprintf("\n");
		}
 */

		if (rbuf[0])
			retry = 0;
	}
	dprintf("\n");

	if (!exit_loop) { /* break by dev I/O error */
		retry++;
		if (!background && retry > retry_count)
			goto bailout;

		sleep(1);
		close(ca_fd);
		SCardDisconnect(hcard, SCARD_RESET_CARD);
		goto repeat;
	}

bailout:
	if (ca_fd >= 0)
		close(ca_fd);

	if (hcard)
		SCardDisconnect(hcard, SCARD_LEAVE_CARD);

	if (cxt) {
		SCardCancel(cxt);
		SCardReleaseContext(cxt);
	}
	return r;
}
