/*
 * descramble TS files via kmod dvb-usb-swbcas (for tesing purpose)
 */
#define _LARGEFILE64_SOURCE
#include <features.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <PCSC/pcsclite.h>
#include <sys/stat.h>
#include <sys/types.h>

#define DEBUG 1

#ifdef DEBUG
#define dprintf(args...) fprintf(stderr, args)
#endif

/* example ATR of BCAS  note: byte[2],[4],[8],[9],[11],[12] may vary. */
static const BYTE bcas_atr[] = {
	0x3B, 0xF0, 0x12, 0x00, 0xFF, 0x91, 0x81,
	0xB1, 0x7C, 0x45, 0x1F, 0x03, 0x99
};
#define BCAS_ATR_LEN 13

static int get_card(SCARDCONTEXT cxt, SCARDHANDLE *h)
{
	LONG ret;
	DWORD len, protocol;
	char *p, *readers;
	DWORD nlen, alen;
	DWORD state;
	BYTE atr[MAX_ATR_SIZE];

	ret = SCardListReaders(cxt, NULL, NULL, &len);
	if (ret != SCARD_S_SUCCESS) {
		fprintf(stderr, "Cannot get length of the IC card list: %s\n",
			pcsc_stringify_error(ret));
		return -1;
	}

	readers = malloc(len);
	if (!readers) {
		perror("Cannot alloc mem for the IC card list.");
		return errno;
	}

	ret = SCardListReaders(cxt, NULL, readers, &len);
	if (ret != SCARD_S_SUCCESS) {
		fprintf(stderr, "Cannot get card list. %s\n",
			pcsc_stringify_error(ret));
		free(readers);
		return -1;
	}

	p = readers;
	while (*p) {
		*h = 0;
		ret = SCardConnect(cxt, p, SCARD_SHARE_SHARED,
				   SCARD_PROTOCOL_T1, h, &protocol);
		if (ret != SCARD_S_SUCCESS) {
			fprintf(stderr,	" cannot connect to [%s]/%s.\n",
				p, pcsc_stringify_error(ret));
			goto next_dev;
		}
/*
		SCardReconnect(h, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1,
				SCARD_RESET_CARD, &protocol); 
		usleep(10*1000);
*/
		alen = sizeof(atr);
		ret = SCardStatus(*h, NULL, &nlen, &state, &protocol,
				  atr, &alen);
		dprintf(" state:0x%02hhx atr-len:%d\n", state, alen);

		{ int i;
		for(i=0; i<alen; i++) dprintf(" %02hhx", atr[i]);
		dprintf("\n");
		}


		if (ret == SCARD_S_SUCCESS && state & SCARD_PRESENT &&
		    alen == BCAS_ATR_LEN &&
		    atr[0] == bcas_atr[0] && atr[1] == bcas_atr[1]) {
			fprintf(stderr, "found a (likely) BCAS card[%s].\n", p);
			free(readers);
			return 0;
		}
		SCardDisconnect(*h, SCARD_RESET_CARD);
	next_dev:
		p += strlen(p) + 1;
	}
	free(readers);

	fprintf(stderr, "cannot find an available BCAS card.\n");
	return -1;
}

extern void sw_bcas_proc_ts(int wfile, uint8_t *buf, size_t len);
extern int dvb_bcas_io_open(void);

SCARDHANDLE  hcard;

int main(int argc, char **argv)
{
	int i;
	int rfile, wfile;
#define BUFSIZE 4096
	uint8_t rbuf[BUFSIZE];
	int dlen;
	
	LONG ret;
	SCARDCONTEXT cxt;

	if (argc < 3) {
		fprintf(stderr, "Usage: %s <src file> <dst file>\n", argv[0]);
		return -1;
	}

	rfile = open(argv[1], O_RDONLY);
	wfile = creat(argv[2], S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
	if (rfile == -1 || wfile == -1) {
		perror("file open error:");
		return -1;
	}

	ret = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &cxt);
	if (ret != SCARD_S_SUCCESS) {
		fprintf(stderr, "Can't connect to pcscd. %s\n",
			pcsc_stringify_error(ret));
		return -1;
	}

	hcard = 0;
	for (i=0; i<5; i++) {
		if (!get_card(cxt, &hcard)) {
			break;
		}
		sleep(5);
	}
	if (i==5) {
		fprintf(stderr, "failed to get a card.\n");
		return -1;
	}

	dvb_bcas_io_open();
	while ((dlen = read(rfile, rbuf, sizeof(rbuf))) > 0) {
		sw_bcas_proc_ts(wfile, rbuf, dlen);
	}


	close(rfile);
	close(wfile);
	if (hcard)
		SCardDisconnect(hcard, SCARD_LEAVE_CARD);

	if (cxt) {
		SCardCancel(cxt);
		SCardReleaseContext(cxt);
	}

	return 0;
}

extern ssize_t dvb_bcas_io_read(char *, size_t);
extern ssize_t dvb_bcas_io_write(const char *, size_t);

void proc_cmd(void)
{
	/* read-(IC Card)-write */

	int r, i;
	LONG ret;
	DWORD protocol;
	DWORD rlen;
	SCARD_IO_REQUEST rpci;

	unsigned char tbuf[256];
	unsigned char rbuf[256];
	unsigned char *p;
	int len;

		/* read {len, seq, APDU[len]} */
		r = dvb_bcas_io_read(tbuf, sizeof(tbuf));
		fprintf(stderr, "read ret:%d t[0]:0x%02hhx\n", r, tbuf[0]);
		if (r < 0) {
			perror("dev read header failed.");
			exit(1);
		}

		if (r < 2) {
			fprintf(stderr, "read ret:%d\n", r);
			exit(1);
		}

		len = tbuf[0] + 2 - r;
		p = tbuf + r;
		while (len > 0) {
			r = dvb_bcas_io_read(p, sizeof(tbuf) - (p - tbuf));
			fprintf(stderr, "read ret:%d\n", r);
			if (r <= 0) {
				perror("dev read data failed.");
				exit(1);
			}
			p += r;
			len -= r;
		}
		dprintf("cmd[%d] read.\n", p - tbuf);

		/* send to the card */
		memcpy(&rpci, SCARD_PCI_T1, sizeof(rpci));
		rlen = sizeof(rbuf) - 2;

		{ int i;
		for(i=0; i < tbuf[0] + 2; i++) dprintf(" %02hhx", tbuf[i]);
		dprintf("\n");
		}

		ret = SCardTransmit(hcard, SCARD_PCI_T1, tbuf + 2, tbuf[0],
				    &rpci, rbuf + 2, &rlen);
		if (ret == SCARD_W_RESET_CARD) {
			ret = SCardReconnect(hcard, SCARD_SHARE_SHARED,
					     SCARD_PROTOCOL_T1,
					     SCARD_LEAVE_CARD, &protocol);
		}

		/* write back to the dev */
		rbuf[1] = tbuf[1];
		if (ret == SCARD_S_SUCCESS) {
			rbuf[0] = rlen & 0xff;
		} else {
			fprintf(stderr, "bcas transmit failed.%s\n",
				pcsc_stringify_error(ret));
			rbuf[0] = 0; /* for reporting error, send empty data */
			SCardDisconnect(hcard, SCARD_RESET_CARD);
			hcard = 0;
			exit(1);
		}

		dprintf("resp(%d) recved\n", rlen);

		ret = dvb_bcas_io_write(rbuf, rbuf[0] + 2);
		if (ret < rbuf[0] + 2) {
			perror("dev write error.");
			exit(1);
		}
		dprintf("wrote back to the dev.\n");
/*
		{ int i;
		for(i=0; i < rbuf[0] + 2; i++) dprintf(" %02hhx", rbuf[i]);
		dprintf("\n");
		}
 */

	return;
}
