/**
 * Software Descrambler Filter for the B-CAS, ISDB-T/S CA System 
 * 
 * (c) 2008 0p1pp1
 *
 * Basesd on code found in gl861.c
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * see Documentation/dvb/README.dvb-usb for more information
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "isdb-psi.h"

/* debug */
#define deb_info(args...) fprintf(stderr, args)
#define deb_err(args...) fprintf(stderr, args)

#define ICCARD_RETCODE_VOID 0
#define ICCARD_RETCODE_OK 0x2100
#define ICCARD_RETCODE_GD_TIER 0x0800
#define ICCARD_RETCODE_GD_PREPPV 0x0400
#define ICCARD_RETCODE_GD_POSTPPV 0x0200
#define ICCARD_RETCODE_PV_PREPPV 0x4480
#define ICCARD_RETCODE_PV_POSTPPV 0x4280
#define ICCARD_TYPE_GENERAL 0x01
#define ICCARD_SW1_OK 0x90
#define ICCARD_SW2_OK 0x00

#define SYS_MNG_ID_B14 0x0301
#define SYS_MNG_ID_B15 0x0201
#define BCAS_PROTO_UNIT 0
#define BCAS_MULTI2_ROUND 4

#define CARD_S_INVALID 0
#define CARD_S_INIT 1
#define CARD_S_READY 2

#define CMD_S_VOID 0
#define CMD_S_QUERYING 1
#define CMD_S_FAILED 2
#define CMD_S_OK 3

struct card_state {
	u32 status;

	u16 cas_id;

	u8 card_vendor; /* [A-Z] */
	u8 card_ver;
	int num_id;
#define BCAS_MAX_ID 8
	u8 id[BCAS_MAX_ID][10]; /* id[0]: card-id, id[1..]: group-ids */
	/* id[][0]: VendorID(ex.'M'), id[][1]:Ver, id[][2-7]:ID, id[8-9]:CC */

	struct demulti2_param param;

	u16 last_card_dirs; /* IC card directions of the last response */
	u16 last_ret_code; /* return code of the last response */

	int lock;

	u8 cmd_seq; /* seq # of the current cmd */

        /*
	 * if error != 0, previous cmd purged out by the new one,
	 *   (the partially-read old cmd was interrupted by the new one)
	 *   and read() will return error on the next call
	 */
	int error; 
	void (*cmd_cb)(struct card_state *bcas, void *arg);
	void *cmd_arg;


	/* cmd header:  buf[0]:payload-len, buf[1]:seq#, buf[2-]:payload */
#define CMD_HEADER_SIZE 2
	/*
	 * cmd_len > 0, cmd_pos == 0,(rsp_pos == 0): cmd request issued
	 * cmd_len > cmd_pos > 0, (rsp_pos == 0): cmd is being read
	 * cmd_len = cmd_pos > 0, (rsp_pos == 0): cmd has been read
	 * cmd_len = cmd_pos > 0, rspbuf[0]+2 > rsp_pos > 0: rsp is being put
	 * cmd_len = cmd_pos > 0, rspbuf[0]+2 = rsp_pos > 0: rsp whole put
	 */
#define MAX_MSG_LEN 268
	int cmd_len;
	int cmd_pos;
	u8 cmdbuf[MAX_MSG_LEN];

	int rsp_pos;
	u8 rspbuf[MAX_MSG_LEN];
};

int proc_emm = 0;
extern void proc_cmd(void);

void writeout_to_file(int wfd, u8 *buf, size_t len)
{
	int ret = write(wfd, buf, len);
	if (ret != len) {
		perror("Output failed.");
		exit(1);
	}
	return;
};

struct sw_bcas_state {
	struct dvb_usb_adapter *adap;

	void (*complete_orig)(int, u8 *, size_t);

        struct dvb_device *dev;
	struct card_state bcas;

	/* for fragmented TS packets at the both ends of the URB data */
	u8 buf_used;
	u8 packet_buf[188];

	struct pat pat;
	struct cat cat;
} State;

static size_t min(size_t a, size_t b)
{
	return (a > b) ? b : a;
}

/*
 * card related funcs
 */

/* call backs for CARD responses (note: caller holds the lock) */
static void sw_bcas_cb_info(struct card_state *bcas, void *dummy);

static void sw_bcas_cb_init(struct card_state *bcas, void *arg)
{
//	wait_queue_head_t *wq = (wait_queue_head_t *)arg;
	static u8 cardinfo_cmd[] = {0x90, 0x32, 0x00, 0x00, 0x00};
	u8 sw1, sw2, ulen, *p = bcas->rspbuf;
	u8 i;

	deb_info("%s called.\n", __func__);

	/* error check (maybe unnecessaly) */

	if (bcas->rsp_pos == 0) {
		bcas->status = CARD_S_INVALID;
		return;
	}

	/* check the response data */
	/* atleast SW1, SW2 should be contained */
	if (bcas->rsp_pos < CMD_HEADER_SIZE + 2 || *p < 2) {
		deb_err("%s: broken response to INIT cmd.\n", __func__);
		bcas->status = CARD_S_INVALID;
		return;
	}

	sw1 = p[CMD_HEADER_SIZE + *p - 2];
	sw2 = p[CMD_HEADER_SIZE + *p - 1];
	deb_info("card init rsp SW1,SW2 = 0x%02x, 0x%02x\n", sw1, sw2);
	if (sw1 != ICCARD_SW1_OK || sw2 != ICCARD_SW2_OK) {
		deb_err("%s: error response[0x%02x%02x] to INIT cmd.\n",
			__func__, sw1, sw2);
		bcas->status = CARD_S_INVALID;
		return;
	}

	p += CMD_HEADER_SIZE;
	ulen = p[1];
	bcas->last_card_dirs = p[2] << 8 | p[3];
	bcas->last_ret_code = p[4] << 8 | p[5];
	bcas->cas_id = p[6] << 8 | p[7];

	if (p[0] != BCAS_PROTO_UNIT || ulen < 57 || bcas->cas_id != ID_CAS_ARIB
	    || bcas->last_ret_code != ICCARD_RETCODE_OK
	    || p[14] != ICCARD_TYPE_GENERAL) {
		deb_err("%s: bad data in the msg. ret:[0x%04x] unit#:0x%02x"
			" len:%d cas:0x%04x type:0x%02x.\n",
		        __func__, bcas->last_ret_code,
			p[0], ulen, bcas->cas_id, p[14]);
		bcas->status = CARD_S_INVALID;
		return;
	}

	/* check if system_management_id[] includes valid IDs */
/*
        q = &p[57];
	for (i=0; i<p[56]; i++) {
		sys_mng_id = q[0] << 8 | q[1];
		if (sys_mng_id == SYS_MNG_ID_B14
		    || sys_mng_id == SYS_MNG_ID_B15)
		     break;
		q += 2;
	}
	if (i == p[56]) {
	        bcas->status = CARD_S_INVALID;
	        return;
	}
*/

	/* copy the initial data */
	bcas->param.round = BCAS_MULTI2_ROUND;
	for (i=0; i<8; i++) {
		bcas->param.k_sys[i] = be32_to_cpup((void *)(p + 16 + i * 4));
	}
	bcas->param.cbc_init.whole = be64_to_cpup((void *)&p[48]);
	
	memcpy(&bcas->id[0][2], &p[8], 6); /* redundant.*/
	bcas->num_id = 1;

	/* set the next cmd */
	memcpy(bcas->cmdbuf + CMD_HEADER_SIZE, cardinfo_cmd,
	       sizeof(cardinfo_cmd));
	bcas->cmdbuf[0] = sizeof(cardinfo_cmd);
	bcas->cmdbuf[1] = ++bcas->cmd_seq;
	bcas->cmd_len = sizeof(cardinfo_cmd) + 2;
	bcas->cmd_pos = 0;
	bcas->cmd_cb = sw_bcas_cb_info;
	bcas->cmd_arg = NULL;
//	wake_up_interruptible_sync(wq);
	bcas->rsp_pos = 0;
	proc_cmd();
	return;
}

static void sw_bcas_cb_info(struct card_state *bcas, void *dummy)
{
	u8 *p = bcas->rspbuf;
	u8 sw1, sw2, ulen;

	deb_info("%s called.\n", __func__);

	/* error check (maybe unnecessaly) */

	if (bcas->rsp_pos == 0) {
		bcas->status = CARD_S_INVALID;
		return;
	}

	/* check the response data */
	if (*p < CMD_HEADER_SIZE + 2) {
		deb_err("%s: broken response.\n", __func__);
		bcas->status = CARD_S_INVALID;
		return;
	}

	sw1 = p[CMD_HEADER_SIZE + *p - 2];
	sw2 = p[CMD_HEADER_SIZE + *p - 1];
	deb_info("card info rsp SW1,SW2 = 0x%02x, 0x%02x\n", sw1, sw2);
	if (sw1 != ICCARD_SW1_OK || sw2 != ICCARD_SW2_OK) {
		deb_err("%s: error response[0x%02x%02x].\n",
			__func__, sw1, sw2);
		bcas->status = CARD_S_INVALID;
		return;
	}

	p += 2;
	ulen = p[1];
	bcas->last_card_dirs = p[2] << 8 | p[3];
	bcas->last_ret_code = p[4] << 8 | p[5];
	bcas->num_id = p[6];
	if (p[0] != BCAS_PROTO_UNIT || ulen < 15 ||
	    bcas->last_ret_code != ICCARD_RETCODE_OK ||
	    bcas->num_id < 1 || bcas->num_id > 8) {
		deb_err("%s: bad data in the msg. ret:[0x%04x],%d IDs.\n",
		        __func__, bcas->last_ret_code, p[6]);
		bcas->status = CARD_S_INVALID;
		return;
	}

	/* copy card IDs */
	memcpy(bcas->id, &p[7], 10 * bcas->num_id);

	bcas->status = CARD_S_READY;
	return;
}


static void sw_bcas_cb_ecm(struct card_state *bcas, void *arg)
{
	struct demulti2_key_info *kinfo = (struct demulti2_key_info *)arg;
	int idx = bcas->rsp_pos;
	unsigned long flags;
	u64 k_odd, k_even;

	deb_info("%s called.\n", __func__);

	kinfo->status = CMD_S_FAILED;

#define ECM_RSP_UNIT_SIZE 21
#define ECM_RSP_SIZE ECM_RSP_UNIT_SIZE + 4

	/* check error rsp */
	if (bcas->rspbuf[0] == 0 || idx != CMD_HEADER_SIZE + ECM_RSP_SIZE) {
		deb_err("%s faild to get a right repsonse.\n", __func__);
		bcas->last_ret_code = ICCARD_RETCODE_VOID;
		return;
	}
	/* check SW1, SW2 */

	if (bcas->rspbuf[idx - 2] != ICCARD_SW1_OK ||
	    bcas->rspbuf[idx - 1] != ICCARD_SW2_OK) {
		deb_err("%s: error response[0x%02x%02x].\n", __func__,
		        bcas->rspbuf[idx - 2], bcas->rspbuf[idx - 1]);
		bcas->last_ret_code = ICCARD_RETCODE_VOID;
		return;
	}

	if (bcas->rspbuf[CMD_HEADER_SIZE] != BCAS_PROTO_UNIT ||
	    bcas->rspbuf[CMD_HEADER_SIZE + 1] != ECM_RSP_UNIT_SIZE) {
		deb_err("%s: bad response.\n", __func__);
		bcas->last_ret_code = ICCARD_RETCODE_VOID;
		return;
	}

	bcas->last_card_dirs = bcas->rspbuf[CMD_HEADER_SIZE + 2] << 8 |
		bcas->rspbuf[CMD_HEADER_SIZE + 3];
	bcas->last_ret_code = bcas->rspbuf[CMD_HEADER_SIZE + 4] << 8 |
		bcas->rspbuf[CMD_HEADER_SIZE + 5];

	if (bcas->last_ret_code != ICCARD_RETCODE_GD_TIER &&
		bcas->last_ret_code != ICCARD_RETCODE_GD_PREPPV &&
		bcas->last_ret_code != ICCARD_RETCODE_GD_POSTPPV &&
		bcas->last_ret_code != ICCARD_RETCODE_PV_PREPPV &&
		bcas->last_ret_code != ICCARD_RETCODE_PV_POSTPPV) {
		deb_err("%s: got error response[0x%04x].\n",
		        __func__, bcas->last_ret_code);
		return;
	}

	/* set k_s & update wrk[][] */
	k_odd = be64_to_cpup((void *)&bcas->rspbuf[CMD_HEADER_SIZE + 6]);
	k_even = be64_to_cpup((void *)&bcas->rspbuf[CMD_HEADER_SIZE + 14]);
//	write_lock_irqsave(&kinfo->lock, flags);
	if (k_odd != kinfo->k_scr[0].whole) {
		deb_info("ECM k_odd updated.\n");
		kinfo->k_scr[0].whole = k_odd;
		core_schedule(kinfo->wrk[0], bcas->param.k_sys,
			      &kinfo->k_scr[0].sub);
	}
	if (k_even != kinfo->k_scr[1].whole) {
		deb_info("ECM k_even updated.\n");
		kinfo->k_scr[1].whole = k_even;
		core_schedule(kinfo->wrk[1], bcas->param.k_sys,
			      &kinfo->k_scr[1].sub);
	}
//	write_unlock_irqrestore(&kinfo->lock, flags);
	kinfo->status = CMD_S_OK;
	return;
}

static void sw_bcas_cb_emm(struct card_state *bcas, void *dummy)
{
	int idx = bcas->rsp_pos;

	deb_info("%s called.\n", __func__);

#define EMM_RSP_UNIT_SIZE 4
#define EMM_RSP_SIZE EMM_RSP_UNIT_SIZE + 4

	/* check error rsp */
	if (bcas->rspbuf[0] == 0 || idx != CMD_HEADER_SIZE + EMM_RSP_SIZE) {
		bcas->last_ret_code = ICCARD_RETCODE_VOID;
		return;
	}
	/* check SW1, SW2 */

	if (bcas->rspbuf[idx - 2] != ICCARD_SW1_OK ||
	    bcas->rspbuf[idx - 1] != ICCARD_SW2_OK) {
		deb_err("%s: error response[0x%02x%02x].\n",
		        __func__, bcas->rspbuf[idx - 2], bcas->rspbuf[idx - 1]);
		bcas->last_ret_code = ICCARD_RETCODE_VOID;
		return;
	}

	if (bcas->rspbuf[CMD_HEADER_SIZE] != BCAS_PROTO_UNIT ||
	    bcas->rspbuf[CMD_HEADER_SIZE + 1] != EMM_RSP_UNIT_SIZE) {
		deb_err("%s: bad response.\n", __func__);
		bcas->last_ret_code = ICCARD_RETCODE_VOID;
		return;
	}
	/* only set status, */
	bcas->last_card_dirs = bcas->rspbuf[CMD_HEADER_SIZE + 2] << 8 |
		bcas->rspbuf[CMD_HEADER_SIZE + 3];

	bcas->last_ret_code = bcas->rspbuf[CMD_HEADER_SIZE + 4] << 8 |
		bcas->rspbuf[CMD_HEADER_SIZE + 5];

	return;
}

static int sw_bcas_set_ecm(u8 *buf, int len, struct demulti2_key_info *kinfo,
			   struct card_state *bcas)
{
	/* ECM cmd: head +  Lc:n, DATA:ecm_data[n], Le:00 */
	/*     (len = 6+n) */
	static u8 ecm_cmd_head[] = {0x90, 0x34, 0x00, 0x00};
	unsigned long flags;
	int plen = 6 + len;
	u8 *p = bcas->cmdbuf;
	struct sw_bcas_state *state = &State;

	deb_info("%s called. len:%d.\n", __func__, len);
	if (plen + 2 > sizeof(bcas->cmdbuf)) {
		deb_err("%s: too long msg(%d).\n", __func__, len);
		return -1;
	}

//	spin_lock_irqsave(&bcas->lock, flags);
	if (bcas->cmd_len > bcas->cmd_pos && bcas->cmd_pos > 0) {
		bcas->error = 0;
		bcas->rspbuf[0] = 0;
		bcas->cmd_cb(bcas, bcas->cmd_arg);
	}
	*p++ = plen & 0xff;
	*p++ = ++bcas->cmd_seq;
	memcpy(p, ecm_cmd_head, sizeof(ecm_cmd_head));
	p += sizeof(ecm_cmd_head);
	*p++ = len;
	memcpy(p, buf, len);
	p += len;
	*p++ = 0;
	bcas->cmd_pos = 0;
	bcas->cmd_len = plen + 2;

	bcas->cmd_cb = sw_bcas_cb_ecm;
	bcas->cmd_arg = kinfo;
//	spin_unlock_irqrestore(&bcas->lock, flags);

//	wake_up_interruptible_sync(&state->dev->wait_queue);
	proc_cmd();
	return 0;
}

static int sw_bcas_set_emm(u8 *buf, struct card_state *bcas)
{
	/* EMM cmd: head +  Lc:n, DATA:emm_data[n], Le:00 */
	/*     (len = 6+n) */
	static u8 emm_cmd_head[] = {0x90, 0x36, 0x00, 0x00};
	u8 len = buf[6] + 7;
	int plen = 6 + len;
	struct sw_bcas_state *state = &State;
	unsigned long flags;
	u8 *p = bcas->cmdbuf;

	if (plen + 2 > sizeof(bcas->cmdbuf))
		return -1;

//	spin_lock_irqsave(&bcas->lock, flags);
	if (bcas->cmd_len > bcas->cmd_pos && bcas->cmd_pos > 0) {
		bcas->error = 0;
		bcas->rspbuf[0] = 0;
		bcas->cmd_cb(bcas, bcas->cmd_arg);
	}
	*p++ = plen && 0xff;
	*p++ = ++bcas->cmd_seq;
	memcpy(p, emm_cmd_head, sizeof(emm_cmd_head));
	p += sizeof(emm_cmd_head);
	*p++ = len;
	memcpy(p, buf, len);
	p += len;
	*p++ = 0;
	bcas->cmd_len = plen + 2;
	bcas->cmd_pos = 0;

	bcas->cmd_cb = sw_bcas_cb_emm;
	bcas->cmd_arg = NULL;
//	spin_unlock_irqrestore(&bcas->lock, flags);

//	wake_up_interruptible_sync(&state->dev->wait_queue);
	proc_cmd();
	return 0;
}


/*
 * TS packet descrambling
 */
static int descramble(struct sw_bcas_state *state, u8 *buf)
{
	int i, j, idx, oe;
	size_t n;
	u8 *p;
	u16 pid = (buf[1] & 0x1f) << 8 | buf[2];
	unsigned long flags;

	if (pid == PID_NONE || !(buf[3] & TS_F_PAYLOAD))
		return 0;

	for (i = 0; i < state->pat.num_prog; i++) {
		/* pid == ECM pid of this program ? */
		if (pid == state->pat.prog[i].pmt.prog_ecm.ca_pid)
			break;

		/* pid == one of the ES of this program? */ 
		for (j = 0; j < state->pat.prog[i].pmt.num_es; j++) {
			if (pid == state->pat.prog[i].pmt.es[j].es_pid)
				break;
		}
		if (j < state->pat.prog[i].pmt.num_es)
			break;
	}
	if (i >= state->pat.num_prog) {
		deb_info("%s: pid:[%04x] is not registered to be descrambled\n",
			 __func__, pid);
		return -1;
	}

	/* use ECM of program[i] */
	idx = state->pat.prog[i].pmt.ecm_idx;
	if (idx == -1) {
		deb_err("%s: cannot find ecm_info for ECM:[%04x]\n",
			__func__, state->pat.prog[i].pmt.prog_ecm.ca_pid);
		return -1;
	}

	if (state->pat.ecmt[idx].pid == PID_NONE
	    || state->pat.ecmt[idx].ver == 0xff
	    || state->pat.ecmt[idx].kinfo.status == CMD_S_VOID
	    || state->pat.ecmt[idx].kinfo.status == CMD_S_FAILED) {
		deb_info("%s: ECM:[%04x](stat:%04x) data not available.\n",
			__func__, state->pat.ecmt[idx].pid,
			state->pat.ecmt[idx].kinfo.status);
		return 1;
	}

	/* oe:0 -> k_odd, oe:1 -> k_even */
	oe = (buf[3] & TS_CA_OE_MASK) ? 0 : 1;
	p = buf + 4;
	if (buf[3] & TS_F_ADAP)
		p += p[0] + 1;
	n = 188 - (p - buf);

//	read_lock_irqsave(&state->pat.ecmt[idx].kinfo.lock, flags);
	demulti2(p, n, &state->bcas.param, state->pat.ecmt[idx].kinfo.wrk[oe]);
//	read_unlock_irqrestore(&state->pat.ecmt[idx].kinfo.lock, flags);

	buf[3] &= TS_CA_REV_MASK;
	return 0;
}


/*
 * PSI related funcs
 */

/* utility funcs */
static void init_psi_state(struct sw_bcas_state *state)
{
	int i, j;

	state->pat.ver = VER_NONE;
	state->pat.ts_id = ID_NONE;
	state->pat.num_prog = 0;
	state->pat.sec.len = 0;
	state->pat.sec.max = sizeof(state->pat.sbuf);
	state->pat.sec.buf = state->pat.sbuf;

	for (i=0; i<ISDB_PSI_MAX_PROG; i++) {
		state->pat.prog[i].prog_id = ID_NONE;
		state->pat.prog[i].pmt_pid = PID_NONE;
		state->pat.prog[i].pmt.ver = VER_NONE;
		state->pat.prog[i].pmt.prog_id = ID_NONE;
		state->pat.prog[i].pmt.pcr_pid = PID_NONE;
		state->pat.prog[i].pmt.prog_ecm.cas_id = ID_NONE;
		state->pat.prog[i].pmt.prog_ecm.ca_pid = PID_NONE;
		state->pat.prog[i].pmt.ecm_ref = state->pat.ecmt;
		state->pat.prog[i].pmt.ecm_idx = -1;

		state->pat.prog[i].pmt.num_es = 0;
		for (j=0; j<ISDB_PSI_MAX_ES; j++) {
			state->pat.prog[i].pmt.es[j].es_pid = PID_NONE;
			state->pat.prog[i].pmt.es[j].es_ecm.cas_id = ID_NONE;
		}
		state->pat.prog[i].pmt.sec.len = 0;
		state->pat.prog[i].pmt.sec.max = sizeof(state->pat.prog[i].pmt.sbuf);
		state->pat.prog[i].pmt.sec.buf = state->pat.prog[i].pmt.sbuf;

		state->pat.ecmt[i].pid = PID_NONE;
		state->pat.ecmt[i].ref_cnt = 0;
		state->pat.ecmt[i].ver = 0xff;
		memset(&state->pat.ecmt[i].kinfo, 0,
		       sizeof(state->pat.ecmt[i].kinfo));
//		rwlock_init(&state->pat.ecmt[i].kinfo.lock);
	}

	state->cat.ver = VER_NONE;
	state->cat.num_cas = 0;
	for (i=0; i<ISDB_PSI_MAX_CA; i++) {
		state->cat.emm[i].ca.cas_id = ID_NONE;
		state->cat.emm[i].ca.ca_pid = PID_NONE;
		state->cat.emm[i].ver = VER_NONE;
		state->cat.emm[i].sec.len = 0;
		state->cat.emm[i].sec.max = sizeof(state->cat.emm[i].sbuf);
		state->cat.emm[i].sec.buf = state->cat.emm[i].sbuf;
	}
	state->cat.sec.len = 0;
	state->cat.sec.max = sizeof(state->cat.sbuf);
	state->cat.sec.buf = state->cat.sbuf;

	return;
}


/* process CA descriptor */
static int set_ca(u8 *b, struct ca *ca, int filter)
{
	u16 id;

	if (b[0] != DESC_CA) return 1;
	id = b[2] << 8 | b[3];
	if (id != ID_CAS_ARIB && filter) return 1;
	ca->cas_id = id;
	ca->ca_pid = (b[4] & 0x1f) << 8 | b[5];
	if (b[1] > 4)
		ca->type = b[6];
	else
		ca->type = 0;

	return 0;
}

/* find an ecm_table[] entry for the program's ECM */
/*  shares an previously used entry for the same ECM-pid, or make a new one */
static void set_ecm_table(struct pmt *pmt)
{
	int i;
	struct ecm_info *ecmt = pmt->ecm_ref;

	if (pmt->prog_ecm.ca_pid == PID_NONE)
		return;

	/* search and reuse the existing one */
	for (i=0; i<ISDB_PSI_MAX_PROG; i++) {
		if (pmt->prog_ecm.ca_pid != ecmt[i].pid)
			continue;

		ecmt[i].ref_cnt++;
		pmt->ecm_idx = i;
		deb_info("%s found the entry[%d]/ref:%d for pid:%04x\n",
			 __func__, i, ecmt[i].ref_cnt, ecmt[i].pid);
		return;
	}

	for (i=0; i<ISDB_PSI_MAX_PROG; i++) {
		if (ecmt[i].ref_cnt > 0)
			continue;

		ecmt[i].ref_cnt = 1;
		ecmt[i].pid = pmt->prog_ecm.ca_pid;
		ecmt[i].ver = 0xff;

		/* should hold the lock? */
		memset(&ecmt[i].kinfo.k_scr, 0, sizeof(ecmt[i].kinfo.k_scr));
		memset(&ecmt[i].kinfo.wrk, 0, sizeof(ecmt[i].kinfo.wrk));
		ecmt[i].kinfo.status = CMD_S_VOID;

		pmt->ecm_idx = i;
		deb_info("%s created a new ECM entry[%d] for pid:%04x\n",
			 __func__, i, ecmt[i].pid);
		return;
	}
	deb_err("ECM table overflowed.\n");
	return;
}


/*
 * psi handlers
 */
static void doECM(u8 *buf, struct ecm_info *info, struct card_state *bcas)
{
	int len;
	u8 ver;
	u8 *p = buf + 4;

	if (buf[3] & TS_F_ADAP)
		p += p[0] + 1;

	if (!(buf[1] & TS_F_PUSI)) {
		deb_err("%s: bad packet.\n", __func__);
		return;
	}
	p += p[0] + 1;

	if (*p != TID_ECM) {
		deb_err("%s: bad table id[%02x].\n", __func__, *p);
		return;
	}

	len = (p[1] & 0x0f) << 8 | p[2];
	len -= 9;  /* -5 for the header(p[3-7]), -4 for tail CRC */
	ver = (p[5] & 0x3e) >> 1;

	if (ver == info->ver) {
/*
		deb_info("%s: same table.\n", __func__);
*/
		return;
	}

	/* check cur/next flag, sec #, last sec # */
	if (len > 176 || !(p[5]&0x01) || p[6] || p[7]) {
		deb_err("%s: bad section.\n", __func__);
		return;
	}

	info->ver = ver;

	/* set the ECM command */
	info->kinfo.status = CMD_S_QUERYING;
	sw_bcas_set_ecm(p+8, len, &info->kinfo, bcas);
	return;
}


static void doEMM(struct secbuf *sec, void *data)
{
	int len, ver;
	u8 *p;
	int i, ret;
	struct emm_info *info = NULL;
	struct card_state *bcas = (struct card_state *)data;

	if (sec->buf[0] != TID_EMM) {
		deb_err("%s: bad table header.\n", __func__);
		return;
	}

	len = (sec->buf[1] & 0x0f) << 8 | sec->buf[2];
	len += 3;
	ver = (sec->buf[5] & 0x3e) >> 1;

	/* check cur/next flag, sec #, last sec # */
	/* based on EMM typeA/B */
	if (!(sec->buf[5]&0x01))
		goto bad_sec;

	p = &sec->buf[8];
	if (p + 13 > sec->buf + len - 4)
		goto bad_sec;

	switch (info->ca.type) {
	case CA_EMM_TYPE_A:
		/* TypeA: multi EMM/sec, no GroupID */
		/* check the 1st EMM (minID) */
		ret = memcmp(&bcas->id[0][2], p, 6);
		if (!ret) {
			sw_bcas_set_emm(p, bcas);
			return;
		}
		if (ret < 0)
			break;

		/* check the second EMM (maxID) */
		p += p[7] + 7;
		if (p + 13 >= sec->buf + len - 4)
			break;

		ret = memcmp(&bcas->id[0][2], p, 6);
		if (!ret) {
			sw_bcas_set_emm(p, bcas);
			return;
		}
		if (ret > 0)
			break;

		p += p[7] + 7;
		while (p + 13 <= sec->buf + len - 4) {
			ret = memcmp(&bcas->id[0][2], p, 6);
			if (!ret) {
				sw_bcas_set_emm(p, bcas);
				return;
			}
			if (ret < 0)
				break;
			p += p[7] + 7;
		}
		break;

	case CA_EMM_TYPE_B:
		/* TypeB: 1 EMM / sec */
		if (sec->buf[6] || sec->buf[7])
			goto bad_sec;

		for (i=0; i < bcas->num_id; i++) {
			ret = memcmp(&bcas->id[i][2], p, 6);
			if (!ret) {
				sw_bcas_set_emm(p, bcas);
				return;
			}
		}
		break;

	default:
		goto bad_sec;
	}

	deb_info("%s: EMM for my card not included.\n", __func__);
	return;

bad_sec:
	deb_err("%s:  bad section.\n", __func__);
	return;
}


static void doCAT(struct secbuf *sec, void *data)
{
	int ver, len, i;
	u8 *b;
	struct cat *cat = data;

	if (sec->buf[0] != TID_CAT) {
		deb_err("%s: bad table header.\n", __func__);
		return;
	}

	len = ((sec->buf[1] & 0x0f) << 8) + sec->buf[2] + 3;
	ver = (sec->buf[5] & 0x3e) >> 1;
	if (ver == cat->ver) {
/*
		deb_info(" same CAT table.%02x\n", ver);
*/
		return;
	}

	if (!(sec->buf[5]&0x01) || sec->buf[6] || sec->buf[7]) {
		deb_err("%s: bad section.\n", __func__);
		return;
	}

	/* new version */
	deb_info("new CAT, ver:%02d len:%d \n", ver, len);
	cat->ver = ver;
	cat->num_cas = 0;
	/* clear old emm */
	for (i=0; i<ISDB_PSI_MAX_CA; i++) {
		cat->emm[i].ca.cas_id = ID_NONE;
		cat->emm[i].ca.ca_pid = PID_NONE;
		cat->emm[i].ver = VER_NONE;
		cat->emm[i].sec.len = 0;
	}
	b = &sec->buf[8];
	while (b < sec->buf + len - 4) {
		if (b[0] == DESC_CA && cat->num_cas < ISDB_PSI_MAX_CA) {
			if ( !set_ca(b, &cat->emm[cat->num_cas].ca, 0) )
				cat->num_cas++;

			deb_info("%s: cas(%04x,%s)/pid:%04x\n", __func__,
				cat->emm[cat->num_cas].ca.cas_id,
				cat->emm[cat->num_cas].ca.type == CA_EMM_TYPE_A ? "a":"b",
				cat->emm[cat->num_cas].ca.ca_pid);
		}
		b += b[1] + 2;
	}
	return;
}



static void doPMT(struct secbuf *sec, void *data)
{
	int ver, len, i, pinfo_len, esinfo_len;
	int skip;
	u16 prog_id;
	u8 *b;
	struct pmt *pmt = data;

	/* TODO: verify CRC? */
	if (sec->buf[0] != TID_PMT) {
		deb_err("%s: bad table header.\n", __func__);
		return;
	}

	len = ((sec->buf[1] & 0x0f) << 8) + sec->buf[2] + 3;
	prog_id = (sec->buf[3] << 8) | sec->buf[4];
	ver = (sec->buf[5] & 0x3e) >> 1;

	if (!(sec->buf[5]&0x01) || sec->buf[6] || sec->buf[7]) {
		deb_err("%s: bad section.\n", __func__);
		return;
	}

	if (ver == pmt->ver && prog_id == pmt->prog_id) {
/*
		deb_info(" same table. prog:%04x %02x\n", prog_id, ver);
*/
		return;
	}

	if (ver == pmt->ver && prog_id != pmt->prog_id) {
		deb_err("%s: bad data (prog:%04x) in PMT table[%04x].\n",
			 __func__, prog_id, pmt->prog_id);
		return;
	}

	/* new version */
	pmt->ver = ver;

	/* clean old data */
	if (pmt->ecm_idx != -1)
		pmt->ecm_ref[pmt->ecm_idx].ref_cnt--;
	pmt->ecm_idx = -1;
	pmt->prog_ecm.ca_pid = PID_NONE;

	pmt->prog_id = prog_id;
	pmt->pcr_pid = (sec->buf[8] & 0x1f) << 8 | sec->buf[9];
	pinfo_len = (sec->buf[10] & 0x0f) << 8 | sec->buf[11];
	b = &sec->buf[12];
	for (i=0; i<pinfo_len; ) {
		if (b[0] == DESC_CA) {
			set_ca(b, &pmt->prog_ecm, 1);
			deb_info("PMT  prog_ecm type:%04x, pid:%04x\n",
				 pmt->prog_ecm.cas_id, pmt->prog_ecm.ca_pid);
			set_ecm_table(pmt);
		}
		i += b[1] + 2;
		b += b[1] + 2;
	}

	pmt->num_es = 0;
	while (b < sec->buf + len - 4) {
		skip = 1;
		if (b[0] == ES_TYPE_MPEG2 || b[0] == ES_TYPE_AAC
		    || b[0] == ES_TYPE_H264 || b[0] == ES_TYPE_TXT) {
			skip = 0;
			pmt->es[pmt->num_es].es_type = b[0];
			pmt->es[pmt->num_es].es_pid = (b[1] & 0x1f)<<8 | b[2];
		}
		esinfo_len = (b[3] & 0x0f) <<8 | b[4];
		deb_info("PMT  es[%04x] type:%02x len:%d\n",
			(b[1]&0x1f)<<8 | b[2], b[0], esinfo_len);
		b+= 5;
		if (skip) {
			b += esinfo_len;
			continue;
		}

		pmt->es[pmt->num_es].es_ecm.cas_id = ID_NONE;
		pmt->es[pmt->num_es].es_ecm.ca_pid = 0x0;

		for (i=0; i<esinfo_len;) {
			switch (b[0]) {
			case DESC_CA:
				set_ca(b, &pmt->es[pmt->num_es].es_ecm, 1);
				if (pmt->es[pmt->num_es].es_ecm.ca_pid != PID_NONE) {
					deb_info("PMT es_ecm.pid !=0x1fff\n");
				}
				break;
			case DESC_STREAM_ID:
				pmt->es[pmt->num_es].tag = b[2];
				break;
			}
			i += b[1] + 2;
			b += b[1] + 2;
		}

		pmt->num_es++;
		if (pmt->num_es >= ISDB_PSI_MAX_ES) break;
	}

	deb_info("new PMT(%04x) created. pcr[%04x] pinfo:%d es:%d ver:%02x\n",
		prog_id, pmt->pcr_pid, pinfo_len, pmt->num_es, ver);

	return;
}


static void doPAT(struct secbuf *sec, void *data)
{
	int ver, len, pnum, i;
	u16 ts_id, prog_id, pid;
	u8 *p;
	struct pmt *pmt;
	struct pat *pat = data;

	/* TODO: check CRC?, using crc32 module like dvb_net.c? */
	if (sec->buf[0] != TID_PAT) {
		deb_err("%s: bad table header.\n", __func__);
		return;
	}

	len = (sec->buf[1] & 0x0f) << 8 | sec->buf[2];
	ts_id = sec->buf[3] << 8 | sec->buf[4];
	ver = (sec->buf[5] & 0x3e) >> 1;
	pnum = (len - 9) / 4;

	/* check cur/next flag, sec #, last sec # */
	if (!(sec->buf[5]&0x01) || sec->buf[6] || sec->buf[7]) {
		deb_err("%s: bad section.\n", __func__);
		return;
	}

	if (ts_id == pat->ts_id && ver == pat->ver) {
/*
		deb_info(" same table. tsid:%04hx %02x\n", ts_id, ver);
 */
		return;
	}

	deb_info("New PAT for ts:%04x, ver:%02x, prog:%d\n", ts_id, ver, pnum);

	pat->ver = ver;
	pat->ts_id = ts_id;

	/* release old PMTs */
	for(i=0; i<pat->num_prog; i++) {
		pat->prog[i].pmt.ver = VER_NONE;
		pat->prog[i].pmt.prog_id = ID_NONE;
		if (pat->prog[i].pmt.ecm_idx != -1) {
			pat->ecmt[pat->prog[i].pmt.ecm_idx].ref_cnt--;
			pat->prog[i].pmt.ecm_idx = -1;
		}
	}

	if (pnum > ISDB_PSI_MAX_PROG)
		pnum = ISDB_PSI_MAX_PROG;

	pat->num_prog = 0;
	p = &sec->buf[8];
	for(i=0; i<pnum; i++) {
		prog_id = (p[0] << 8) | p[1];
		pid = (p[2] << 8 | p[3]) & 0x1fff;
		p += 4;
		if (!prog_id)
			continue;
		pat->prog[pat->num_prog].prog_id = prog_id;
		pat->prog[pat->num_prog].pmt_pid = pid;

		pmt = &pat->prog[pat->num_prog].pmt;
		pmt->ver = VER_NONE;
		pmt->prog_id = prog_id;
		pmt->ecm_idx = -1;
		pmt->sec.len = 0;

		pat->num_prog++;
		deb_info("prog:%04x, pmt_pid:%04x\n", prog_id, pid);
	}

	return;
}

/* section parser */
static void doSection(u8 *buf, struct secbuf *sec,
		      void (*cb)(struct secbuf *, void *), void *data)
{
	u8 *b = &buf[4];
	int len = 184;
	int nlen = 0;

	/* skip adaptation field */
	if (buf[3] & TS_F_ADAP) {
		len -= buf[4]+1;
		b += buf[4]+1;
	}

	/* check PUSI flag */
	if (buf[1] & TS_F_PUSI) {
		nlen = len - *b - 1;
		len = *b;
		b++;
	}

	/* fill the previous section */
	sec->cc++;
	sec->cc &= TS_CC_MASK;
	if (len && sec->len && (buf[3] & TS_CC_MASK) == sec->cc) {
		if (len < sec->len - sec->cur) { /* not the end */
			memcpy(sec->buf + sec->cur, b, len);
			sec->cur += len;
			if (nlen > 0)
				deb_err("%s: broken section.\n", __func__);
		} else {
			memcpy(sec->buf + sec->cur, b, sec->len - sec->cur);
//			deb_info("%s callback called.\n", __func__);
			cb(sec, data);
			if (nlen && (len > sec->len - sec->cur))
				deb_err("%s:illegal section gap.\n", __func__);
		}
	}

	/* start  a new section */
	b += len;
	while (nlen>=3) {
		if ( b[0] == 0xff ) break;

		sec->len =  ((b[1] & 0x0f) << 8 ) + b[2] + 3;
		sec->cur = 0;
		sec->cc = buf[3] & 0x0f;
		deb_info("new sec. pid:%04x tid:%02x len:%d cc:%01x\n",
			((buf[1]&0x1f)<<8)+buf[2],  b[0], sec->len, sec->cc);
		if (sec->len > sec->max) {
			deb_err("too long(>= %d) section.\n", sec->max);
			sec->len = 0;
			break;
		}
		if (nlen < sec->len) {
			memcpy(sec->buf, b, nlen);
			sec->cur = nlen;
			break;
		}
		memcpy(sec->buf, b, sec->len);
		deb_info("%s callback called.\n", __func__);
		cb(sec, data);
		nlen -= sec->len;
		b += sec->len;
	}
	
	return;
}


/*
 * process each TS packet
 */
static void proc_packet(struct sw_bcas_state *state, u8 *buf)
{
	int i, idx, ret;
	u16 pid = (buf[1]&0x1f) << 8 | buf[2];

	if (!(buf[3] & TS_F_PAYLOAD) || pid == PID_NONE)
		return;

	/* descrabmle 1st */
	if (buf[3] & TS_CA_MASK && state->bcas.status & CARD_S_READY) {
		ret = descramble(state, buf);
		if (ret < 0) {
//			deb_err("%s: failed to descrabmle.\n", __func__);
			return;
		}
	}

	buf[3] &= TS_CA_REV_MASK;

	/* proc PSI */
	if (pid == PID_PAT) {
		doSection(buf, &state->pat.sec, doPAT, &state->pat);
		return;
	} else if (pid == PID_CAT) {
		doSection(buf, &state->cat.sec, doCAT, &state->cat);
		return;
	}

	for (i=0; i<state->pat.num_prog; i++) {
		if (pid == state->pat.prog[i].pmt_pid) {
			doSection(buf, &state->pat.prog[i].pmt.sec,
				  doPMT, &state->pat.prog[i].pmt);
			return;
		}
		idx = state->pat.prog[i].pmt.ecm_idx;
		if (state->bcas.status == CARD_S_READY &&
		    state->bcas.cas_id == state->pat.prog[i].pmt.prog_ecm.cas_id
		    && idx != -1 && pid == state->pat.ecmt[idx].pid) {
		        doECM(buf, &state->pat.ecmt[idx], &state->bcas);
			return;
		}
	}

	if (proc_emm && state->bcas.status == CARD_S_READY) {
		for (i=0; i<state->cat.num_cas; i++) {
			if (state->bcas.cas_id == state->cat.emm[i].ca.cas_id
			    && pid == state->cat.emm[i].ca.ca_pid) {
				doSection(buf, &state->cat.emm[i].sec,
					  doEMM, &state->bcas);
				return;
			}
		}
	}
	return;
}


/*
 * wrapper func. of the URB data handler
 */
void sw_bcas_proc_ts(int wfile, u8 *buf, size_t len)
{
	u8 *p = buf;
	struct sw_bcas_state *state = &State;
	u8 *p1;

	/* TODO: check card's async-response/status-change here? */
	if (!state || !state->complete_orig)
		return;

	/* cut out one TS packet */

	if (state->buf_used) {
		if (len < 188 - state->buf_used) {
			memcpy(&state->packet_buf[state->buf_used], buf, len);
			state->buf_used += len;
			goto bailout;
		}
		memcpy(&state->packet_buf[state->buf_used], buf,
		       188 - state->buf_used);
		proc_packet(state, state->packet_buf);
		state->complete_orig(wfile, state->packet_buf, 188);
		p += 188 - state->buf_used;
		state->buf_used = 0;
	}

	p1 = p;
	while (p  < buf + len) {
		if (*p != 0x47) {
			p++;
			continue;
		}
		if (p + 188 > buf + len)
			break;
		proc_packet(state, p);
		p += 188;
	}
	if (p1 != p)
		state->complete_orig(wfile, p1, p - p1);

	/* save the remaining data */
	if (p < buf + len) {
		state->buf_used = buf + len - p;
		memcpy(state->packet_buf, p, state->buf_used);
	}

	/* call the original handler (= dvb_usb_data_complete()) */
bailout:
//	state->complete_orig(wfile, buf, len);
	return;
}

/*
 * fops
 */
ssize_t dvb_bcas_io_write(const char *buf, size_t count)
{
	struct sw_bcas_state *state = &State;
	struct card_state *bcas = &state->bcas;
	unsigned long flags;
	char *p;

	deb_info("%s\n", __func__);

//	spin_lock_irqsave(&bcas->lock, flags);

	/* if cmd not read out yet */
	if (bcas->cmd_pos != bcas->cmd_len || bcas->cmd_len == 0) {
//		spin_unlock_irqrestore(&bcas->lock, flags);
		return -1;
	}

	if (count + bcas->rsp_pos > sizeof(bcas->rspbuf)) {
		bcas->rsp_pos = 0;
//		spin_unlock_irqrestore(&bcas->lock, flags);
		return -1;
	}

	p = bcas->rspbuf + bcas->rsp_pos;
	bcas->rsp_pos += count;
//	spin_unlock_irqrestore(&bcas->lock, flags);

	/* copy_from_user() can sleep, so release the spinlock temporalily */
	memcpy(p, buf, count);

//	spin_lock_irqsave(&bcas->lock, flags);
	if (bcas->rsp_pos > bcas->rspbuf[0] + CMD_HEADER_SIZE) {
		bcas->rsp_pos = 0;
//		spin_unlock_irqrestore(&bcas->lock, flags);
		return -1;
	}

	if (bcas->rsp_pos == bcas->rspbuf[0] + CMD_HEADER_SIZE) {
		/* whole rsp received */
		/* check seq # first. */
		if (bcas->rspbuf[1] == bcas->cmdbuf[1]) {
			/* clear the cmdbuf */
			bcas->cmd_len = 0;
			bcas->cmd_pos = 0;
			bcas->error = 0;

			/* call cb */
			if (bcas->cmd_cb) {
				deb_info("%s:calling a call-back func.\n", __func__);
				bcas->cmd_cb(bcas, bcas->cmd_arg);
			}
		}
		deb_info("%s: finished proc'ing a card resp.\n", __func__);
		/* clear the rspbuf */
		bcas->rsp_pos = 0;
	}

//	spin_unlock_irqrestore(&bcas->lock, flags);
	return count;
}

ssize_t dvb_bcas_io_read(char *buf, size_t count)
{
	struct sw_bcas_state *state = &State;
	struct card_state *bcas = &state->bcas;
	unsigned long flags;
	char *p;

	deb_info("%s\n", __func__);

//	spin_lock_irqsave(&bcas->lock, flags);

	if (bcas->cmd_pos >= bcas->cmd_len && bcas->error == 0) {
		deb_info("%s: no data, going to sleep.\n", __func__);
		return -1;
	}

	if (bcas->error) {
		int ret = bcas->error;
		bcas->error = 0;
		bcas->rspbuf[0] = 0; /* set rsp-error */
		/* call cb */
		if (bcas->cmd_cb)
			bcas->cmd_cb(bcas, bcas->cmd_arg);
		bcas->rsp_pos = 0;
//		spin_unlock_irqrestore(&bcas->lock, flags);
		return ret;
	}

	count = min(count, (size_t)(bcas->cmd_len - bcas->cmd_pos));
	p = bcas->cmdbuf + bcas->cmd_pos;
	bcas->cmd_pos += count;

	memcpy(buf, p, count);

	return count;
}

void init_state(void)
{
	memset(&State, 0, sizeof(State));
	State.complete_orig = writeout_to_file;
	init_psi_state(&State);
	State.bcas.status = CARD_S_INVALID;
	return;
}

int dvb_bcas_io_open(void)
{
	struct sw_bcas_state *state = &State;
	struct card_state *bcas = &state->bcas;

	int ret=0;
	static u8 init_cmd[] = {0x90, 0x30, 0x00, 0x00, 0x00};

	deb_info("%s\n", __func__);

	/* lock ? */

	init_state();
	bcas->last_card_dirs = 0;
	bcas->last_ret_code = 0;

	bcas->cmd_seq = 0;
	bcas->error = 0;
	bcas->cmd_cb = NULL;
	bcas->cmd_arg = NULL;
	bcas->cmd_len = 0;
	bcas->cmd_pos = 0;
	bcas->rsp_pos = 0;

	/* start initializing */
	bcas->status = CARD_S_INIT;
	/*  cmd data: {len, seq, apd[len]} */
	bcas->cmdbuf[0] = sizeof(init_cmd);
	bcas->cmdbuf[1] = ++bcas->cmd_seq;
	memcpy(bcas->cmdbuf + CMD_HEADER_SIZE, init_cmd, sizeof(init_cmd));
	bcas->cmd_len = sizeof(init_cmd) + CMD_HEADER_SIZE;
	bcas->cmd_cb = sw_bcas_cb_init;
//	bcas->cmd_arg = &dvbdev->wait_queue;
	bcas->cmd_arg = NULL;

	proc_cmd();
	return ret;
}

