/**
 * Driver for Friio ISDB-T/S USB2.0 devices.
 * 
 * (c) 2008 0p1pp1
 *
 * Basesd on code found in gl861.c
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * see Documentation/dvb/README.dvb-usb for more information
 */

/**
 *      Components of Friio
 *       USB hub:                                AU4254
 *         USB controller(+ TS dmx & streaming): GL861
 *         Frontend:                             comtech JDVBT-90502
 *             (tuner PLL:                       tua6034, I2C addr:(0xC0 >> 1))
 *             (OFDM demodulator:                TC90502, I2C addr:(0x30 >> 1))
 *         LED x3 controll:                      PIC 16F676
 *         EEPROM:                               24C08
 *
 *        (USB smart card reader:                AU9522)
 *
 */

#include "friio.h"

#include "jdvbt90502.h"
#include "sw-bcas.h"

/* debug */
static int dvb_usb_friio_debug;
module_param_named(debug,dvb_usb_friio_debug, int, 0644);
MODULE_PARM_DESC(debug, "set debugging level bit flags(1=info,2=xfer)." DVB_USB_DEBUG_STATUS);

DVB_DEFINE_MOD_OPT_ADAPTER_NR(adapter_nr);

#define deb_info(args...) dprintk(dvb_usb_friio_debug, 0x01, args)
#define deb_xfer(args...) dprintk(dvb_usb_friio_debug, 0x02, args)

/* use bulk */
#undef FRIIO_USE_ISOC

/**
 * for the indirect I2C access to the PLL via FE,
 * whole I2C protocol data to the PLL must be sent to the FE's I2C register.
 * so we have to use control msg with data followed, and
 * another USB request number is provided for that purpose.
 *
 * this func sends wbuf[1..] to the I2C register wbuf[0] at addr (= at FE).
 * TODO: consider using another i2c_algorithm/i2c_adapter for PLL
 */
static int gl861_i2c_ctrlmsg_data(struct dvb_usb_device *d, u8 addr,
			     u8 *wbuf, u16 wlen, u8 *rbuf, u16 rlen)
{
	u16 index = wbuf[0]; /* fixed to JDVBT90502_2ND_I2C_REG(0xFE) */
	u16 value = addr << (8 + 1);
	int wo = (rbuf == NULL || rlen == 0); /* write only */
	u8 req, type;

	deb_xfer("write to PLL:0x%02x via FE reg:0x%02x, len:%d\n",
		 wbuf[1], wbuf[0], wlen - 1);

	if (wo && wlen >= 2) {
		req = GL861_REQ_I2C_DATA_CTRL_WRITE;
		type = GL861_WRITE;
		udelay(20);
		return usb_control_msg(d->udev, usb_sndctrlpipe(d->udev, 0),
				       req, type, value, index,
				       &wbuf[1], wlen - 1, 2000);
	}

	deb_xfer("not supported ctrl-msg, aborting.");
	return -EINVAL;
}

static struct jdvbt90502_config friio_jdvbt90502_config;

static int gl861_i2c_msg(struct dvb_usb_device *d, u8 addr,
			 u8 *wbuf, u16 wlen, u8 *rbuf, u16 rlen)
{
	u16 index;
	u16 value = addr << (8 + 1);
	int wo = (rbuf == NULL || rlen == 0); /* write-only */
	u8 req, type;
	unsigned int pipe;

	/* special case for the indirect I2C access to the PLL via FE, */
	if (addr == friio_jdvbt90502_config.demod_address &&
	    wbuf[0] == JDVBT90502_2ND_I2C_REG)
		return gl861_i2c_ctrlmsg_data(d, addr, wbuf, wlen, rbuf, rlen);

	if (wo) {
		req = GL861_REQ_I2C_WRITE;
		type = GL861_WRITE;
		pipe = usb_sndctrlpipe(d->udev, 0);
	} else { /* rw */
		req = GL861_REQ_I2C_READ;
		type = GL861_READ;
		pipe = usb_rcvctrlpipe(d->udev, 0);
	}

	switch (wlen) {
	case 1:
		index = wbuf[0];
		break;
	case 2:
		index = wbuf[0];
		value = value + wbuf[1];
		break;
	case 3:
		/* special case for 16bit register-address */
		index = (wbuf[2] << 8) | wbuf[0];
		value = value + wbuf[1];
		break;
	default:
		deb_xfer("wlen = %x, aborting.", wlen);
		return -EINVAL;
	}
	udelay(20);
	return usb_control_msg(d->udev, pipe, req, type,
			       value, index, rbuf, rlen, 2000);
}

/* I2C */
static int gl861_i2c_xfer(struct i2c_adapter *adap, struct i2c_msg msg[],
			  int num)
{
	struct dvb_usb_device *d = i2c_get_adapdata(adap);
	int i;


	if (num > 2)
		return -EINVAL;

	if (mutex_lock_interruptible(&d->i2c_mutex) < 0)
		return -EAGAIN;

	for (i = 0; i < num; i++) {
		/* write/read request */
		if (i+1 < num && (msg[i+1].flags & I2C_M_RD)) {
			if (gl861_i2c_msg(d, msg[i].addr, msg[i].buf,
					  msg[i].len, msg[i+1].buf, msg[i+1].len) < 0)
				break;
			i++;
		} else
			if (gl861_i2c_msg(d, msg[i].addr, msg[i].buf,
					  msg[i].len, NULL, 0) < 0)
				break;
	}

	mutex_unlock(&d->i2c_mutex);
	return i;
}

static u32 gl861_i2c_func(struct i2c_adapter *adapter)
{
	return I2C_FUNC_I2C;
}

static struct i2c_algorithm gl861_i2c_algo = {
	.master_xfer   = gl861_i2c_xfer,
	.functionality = gl861_i2c_func,
};

/* Callbacks for DVB USB */
static int gl861_identify_state(struct usb_device *udev,
				struct dvb_usb_device_properties *props,
				struct dvb_usb_device_description **desc,
				int *cold)
{
	*cold = 0;
	return 0;
}

#define FRIIO_DEMOD_ADDR  (0x30 >> 1)
#define FRIIO_PLL_ADDR  (0xC0 >> 1)

static struct jdvbt90502_config friio_jdvbt90502_config = {
	.demod_address = FRIIO_DEMOD_ADDR,
	.pll_address = FRIIO_PLL_ADDR,
};

static int friio_initialize(struct dvb_usb_device *d);

static int friio_frontend_attach(struct dvb_usb_adapter *adap)
{
	if (friio_initialize(adap->dev) < 0)
		return -EIO;

	adap->fe = dvb_attach(jdvbt90502_attach, &friio_jdvbt90502_config,
			      &adap->dev->i2c_adap);
	if (adap->fe == NULL)
		return -EIO;

	return 0;
}

/* light the green LED & do some streaming cmds */
static u8 strm_start_cmds[] = {
	0x0B, 0x0F, 0x03, 0x07,  0x03, 0x07, 0x0B, 0x0F,
	0x0B, 0x0F, 0x03, 0x07,  0x03, 0x07, 0x0B, 0x0F,
	0x03, 0x07, 0x03, 0x07,  0x03, 0x07, 0x03, 0x07,
	0x03, 0x07, 0x03, 0x07,  0x03, 0x07, 0x03, 0x07,
	0x03, 0x07, 0x03, 0x07,  0x0B, 0x0F, 0x0B, 0x0F,
	0x0B, 0x0F, 0x0B, 0x0F,  0x0B, 0x0F, 0x0B, 0x0F,
	0x0B, 0x0F, 0x0B, 0x0F,  0x03, 0x07, 0x0B, 0x0F,
	0x0B, 0x0F, 0x03, 0x07,  0x03, 0x07, 0x0B, 0x0F,
	0x03, 0x07, 0x03, 0x07,  0x01, 0x05
};
const static int strm_start_cmd_len = sizeof(strm_start_cmds);

/* light the purple LED & do some streaming cmds(?) */
static u8 strm_stop_cmds[] = {
	0x0B, 0x0F, 0x03, 0x07,  0x0B, 0x0F, 0x03, 0x07,
	0x03, 0x07, 0x0B, 0x0F,  0x03, 0x07, 0x0B, 0x0F,
	0x0B, 0x0F, 0x03, 0x07,  0x0B, 0x0F, 0x0B, 0x0F,
	0x0B, 0x0F, 0x0B, 0x0F,  0x0B, 0x0F, 0x0B, 0x0F,
	0x0B, 0x0F, 0x0B, 0x0F,  0x03, 0x07, 0x03, 0x07,
	0x03, 0x07, 0x03, 0x07,  0x03, 0x07, 0x03, 0x07,
	0x03, 0x07, 0x03, 0x07,  0x0B, 0x0F, 0x0B, 0x0F,
	0x0B, 0x0F, 0x0B, 0x0F,  0x0B, 0x0F, 0x0B, 0x0F,
	0x0B, 0x0F, 0x0B, 0x0F,  0x09, 0x0D, 0x01
};
const static int strm_stop_cmd_len = sizeof(strm_stop_cmds);

static int friio_streaming_ctrl(struct dvb_usb_adapter *adap, int onoff)
{
	int i;
	int ret;
	struct i2c_msg msg;
	u8 *cmdlist;
	int len;
	u8 buf[2];

	deb_info("%s called.(%d)\n", __func__, onoff);

	buf[0] = 0x00;

	msg.addr = 0x00;
	msg.flags = 0;
	msg.len = 2;
	msg.buf = buf;

	if (onoff) {
		cmdlist = strm_start_cmds;
		len = strm_start_cmd_len;
	} else {
		cmdlist = strm_stop_cmds;
		len = strm_stop_cmd_len;
	}

	for (i = 0; i < len; i++) {
		buf[1] = cmdlist[i];
		ret = gl861_i2c_xfer(&adap->dev->i2c_adap, &msg, 1);
		if (ret != 1) {
			deb_info("%s cmd[%d]: ret == %d\n", __func__, i, ret);
			return -EREMOTEIO;
		}
		msleep(1);
	}
	return 0;
}


/* TODO: move these init cmds to the FE's init routine? */
static u8 streaming_init_cmds[][2] = {
	{0x33, 0x08},
	{0x37, 0x40},
	{0x3A, 0x1F},
	{0x3B, 0xFF},
	{0x3C, 0x1F},
	{0x3D, 0xFF},
	{0x38, 0x00},
	{0x35, 0x00},
	{0x39, 0x00},
	{0x36, 0x00},
};
static int cmdlen = sizeof(streaming_init_cmds) / 2;


static int friio_initialize(struct dvb_usb_device *d)
{
	int ret, i, retry = 0; 
	u8 rbuf;
	u8 wbuf[3];

	deb_info("%s called.\n", __func__);

	/* use gl861_i2c_msg instead of gl861_i2c_xfer(), */
	/* because the i2c device is not set up yet. */	
	wbuf[0] = 0x11;
	wbuf[1] = 0x02;
	ret = gl861_i2c_msg(d, 0x00, wbuf, 2, NULL, 0);
	if (ret < 0)
		goto error;
	msleep(2);

	wbuf[0] = 0x11;
	wbuf[1] = 0x00;
	ret = gl861_i2c_msg(d, 0x00, wbuf, 2, NULL, 0);
	if (ret < 0)
		goto error;
	msleep(1);

restart:
	/* TODO: following msgs should be in the FE's init code? */ 
	/* ============ start DEMOD cmds ================== */
	/* read PLL status to clear the POR bit */
	wbuf[0] = JDVBT90502_2ND_I2C_REG;
	wbuf[1] = (FRIIO_PLL_ADDR << 1) + 1; /* +1 for reading */
	ret = gl861_i2c_msg(d, FRIIO_DEMOD_ADDR, wbuf, 2, NULL, 0);
	if (ret < 0)
		goto error;

	msleep(5);
	/* note: DEMODULATOR has 16bit register-address. */
	wbuf[0] = 0x00;
	wbuf[2] = 0x01; /* reg addr: 0x0100 */
	wbuf[1] = 0x00; /* val: not used */
	ret = gl861_i2c_msg(d, FRIIO_DEMOD_ADDR, wbuf, 3, &rbuf, 1); 
	if (ret < 0)
		goto error;
/*
	msleep(1);
	wbuf[0] = 0x80;
	wbuf[1] = 0x00;
	ret = gl861_i2c_msg(d, FRIIO_DEMOD_ADDR, wbuf, 2, &rbuf, 1);
	if (ret < 0)
		goto error;
 */
	if (rbuf & 0x80) {  /* still in POR? */
		if (++retry > 3) {
			deb_info("failed to get the correct FE demod status:0x%02x\n", rbuf);
			goto error;
		}
		msleep(100);
		goto restart;
	}

	/* TODO: check return value in rbuf */
	/* =========== end DEMOD cmds ===================== */
	msleep(1);

	wbuf[0] = 0x30;
	wbuf[1] = 0x04;
	ret = gl861_i2c_msg(d, 0x00, wbuf, 2, NULL, 0);
	if (ret < 0)
		goto error;

	msleep(2);
	/* TODO: following 2 cmds unnecessary? */
	wbuf[0] = 0x00;
	wbuf[1] = 0x01;
	ret = gl861_i2c_msg(d, 0x00, wbuf, 2, NULL, 0);
	if (ret < 0)
		goto error;

	wbuf[0] = 0x06;
	wbuf[1] = 0x0F;
	ret = gl861_i2c_msg(d, 0x00, wbuf, 2, NULL, 0);
	if (ret < 0)
		goto error;

	/* some streaming cmds (maybe) */
	msleep(10);
	for (i = 0; i<cmdlen; i++) {
		ret = gl861_i2c_msg(d, 0x00, streaming_init_cmds[i], 2,
				    NULL, 0);
		if (ret < 0)
			goto error;
		msleep(1);
	}
	msleep(20);

	/* purple LED on & some streaming ctl */
	ret = friio_streaming_ctrl(&d->adapter[0], 0);
	if (ret < 0)
		goto error;

	return 0;

error:
	deb_info("%s:ret == %d\n", __func__, ret);
	return -EIO;
}


/* DVB USB Driver stuff */
static void friio_disconnect(struct usb_interface *intf)
{
	struct dvb_usb_device *d = usb_get_intfdata(intf);
	if (d)
		sw_bcas_detach(&d->adapter[0]);

	dvb_usb_device_exit(intf);
	return;
}

static struct dvb_usb_device_properties friio_properties;

static int friio_probe(struct usb_interface *intf,
		       const struct usb_device_id *id)
{
	struct dvb_usb_device *d;
	struct usb_host_interface *alt;
	int ret;

	if (intf->num_altsetting < GL861_ALTSETTING_COUNT)
		return -ENODEV;

#ifdef FRIIO_USE_ISOC
	alt = usb_altnum_to_altsetting(intf, FRIIO_ISOC_ALTSETTING);
#else
	alt = usb_altnum_to_altsetting(intf, FRIIO_BULK_ALTSETTING);
#endif
	if (alt == NULL) {
		deb_rc("not alt found!\n");
		return -ENODEV;
	}

	ret = usb_set_interface(interface_to_usbdev(intf),
				alt->desc.bInterfaceNumber,
				alt->desc.bAlternateSetting);
	if (ret != 0) {
		deb_rc("failed to set alt-setting!\n");
		return ret;
	}

	ret = dvb_usb_device_init(intf, &friio_properties,
					THIS_MODULE, &d, adapter_nr);

	if (ret == 0) {
		sw_bcas_attach(&d->adapter[0]);
		friio_streaming_ctrl(&d->adapter[0], 1);
	}

	return ret;
}


static struct usb_device_id friio_table [] = {
		{ USB_DEVICE(USB_VID_774, USB_PID_FRIIO_WHITE) },
		{ }		/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, friio_table);


static struct dvb_usb_device_properties friio_properties = {
	.caps = DVB_USB_IS_AN_I2C_ADAPTER,
	.usb_ctrl = DEVICE_SPECIFIC,

	/* NOTE: (void *)priv in struct dvb_usb_adapter points to */
	/* a struct sw_bcas_state which is allocated by the sw_bcas module */
	.size_of_priv     = 0,

	.identify_state   = gl861_identify_state,

	.num_adapters = 1,
	.adapter = {{
/**
 * caps:0 =>  no pid filter, 188B TS packet
 *
 * GL861 should have a H/W pid filter, but I don't know how to controll it.
 * To enable the feature, set ".caps" to 
 *    DVB_USB_ADAP_HAS_PID_FILTER | DVB_USB_ADAP_PID_FILTER_CAN_BE_TURNED_OFF,
 *  and set .pid_filter_count, .pid_filter_ctrl and .pid_filter appropriately.
 */
		.caps  = 0,

		.frontend_attach  = friio_frontend_attach,
		.streaming_ctrl = friio_streaming_ctrl,

		.stream = {
#ifdef FRIIO_USE_ISOC
			.type = USB_ISOC,
#else
			.type = USB_BULK,
#endif
			.count = 8, /* <= MAX_NO_URBS_FOR_DATA_STREAM = 10 */
			.endpoint = 0x01,
			.u = {
#ifdef FRIIO_USE_ISOC
				.isoc = {
					/* bit rate: 940[B] @ 4kHz = 30Mbps */
					/* buf size: 16 x 940 = 15kB/URB */
					/* req rate: 15kB/16Mbps = 7.6ms/URB */
					.framesperurb = 16,
					.framesize = 940,
					.interval = 2, /* *[125us] 8khz */
				}
#else
				/* GL861 has 6KB buf inside */
				.bulk = {
					.buffersize = 16384,
				}
#endif
			}
		},
	}},
	.i2c_algo         = &gl861_i2c_algo,

	.num_device_descs = 1,
	.devices = {
		{   "774 Friio ISDB-T USB2.0",
			{ &friio_table[0], NULL },
			{ NULL },
		},
	}
};

static struct usb_driver friio_driver = {
	.name		= "dvb_usb_friio",
	.probe		= friio_probe,
	.disconnect	= friio_disconnect,
	.id_table	= friio_table,
};


/* module stuff */
static int __init friio_module_init(void)
{
	int ret;

	if ((ret = usb_register(&friio_driver))) {
		err("usb_register failed. Error number %d", ret);
		return ret;
	}

	return 0;
}


static void __exit friio_module_exit(void)
{
	/* deregister this driver from the USB subsystem */
	usb_deregister(&friio_driver);
}

module_init (friio_module_init);
module_exit (friio_module_exit);

MODULE_AUTHOR("0p1pp1");
MODULE_DESCRIPTION("Driver for Friio ISDB-T USB2.0 Receiver");
MODULE_VERSION("0.1");
MODULE_LICENSE("GPL");
